# -*- coding: utf-8 -*-
import ast
import builtins
from contextlib import contextmanager
from enum import Enum
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, GetAttrSubSymbols
from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
from nbsafety.data_model.scope import NamespaceScope


class MethodSpecialCase(Enum):
    normal = 'normal'
    list_append = 'list_append'


if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Set, Tuple, Union
    from ..safety import NotebookSafety
    from nbsafety.data_model.scope import Scope
    SymbolRef = Union[str, AttrSubSymbolChain]
    AttrSubVal = Union[str, int]
    DeepRef = Tuple[int, Optional[str], Tuple[SymbolRef, ...]]
    Mutation = Tuple[int, Tuple[SymbolRef, ...], MethodSpecialCase]
    RefCandidate = Optional[Tuple[int, int, Optional[str]]]
    RecordedArgs = Set[SymbolRef]
    DeepRefCandidate = Tuple[RefCandidate, MethodSpecialCase, RecordedArgs]
    SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
    TextualCallNestingStackFrame = Tuple[Scope, bool]
    TextualCallNestingStack = List[TextualCallNestingStackFrame]
    AttrSubStackFrame = Tuple[
        List[SavedStoreData],
        Set[DeepRef],
        Set[Mutation],
        List[DeepRefCandidate],
        Scope,
        Scope,
        TextualCallNestingStack,
        bool,
        List[bool],
        NamespaceScope
    ]
    AttrSubStack = List[AttrSubStackFrame]

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AttrSubTracingManager(object):
    def __init__(self, safety: 'NotebookSafety',
                 active_scope: 'Scope', trace_event_counter: 'List[int]'):
        self.safety = safety
        self.original_active_scope = active_scope
        self.active_scope = active_scope
        self.should_record_args = False
        self.trace_event_counter = trace_event_counter
        self.attrsub_tracer_name = '_NBSAFETY_ATTR_TRACER'
        self.end_tracer_name = '_NBSAFETY_ATTR_TRACER_END'
        self.arg_recorder_name = '_NBSAFETY_ARG_RECORDER'
        self.scope_pusher_name = '_NBSAFETY_SCOPE_PUSHER'
        self.scope_popper_name = '_NBSAFETY_SCOPE_POPPER'
        self.literal_tracer_name = '_NBSAFETY_LITERAL_TRACER'
        setattr(builtins, self.attrsub_tracer_name, self.attrsub_tracer)
        setattr(builtins, self.end_tracer_name, self.end_tracer)
        setattr(builtins, self.arg_recorder_name, self.arg_recorder)
        setattr(builtins, self.scope_pusher_name, self.scope_pusher)
        setattr(builtins, self.scope_popper_name, self.scope_popper)
        setattr(builtins, self.literal_tracer_name, self.literal_tracer)
        self.ast_transformer = AttrSubTracingNodeTransformer(
            self.attrsub_tracer_name,
            self.end_tracer_name,
            self.arg_recorder_name,
            self.scope_pusher_name,
            self.scope_popper_name,
            self.literal_tracer_name
        )
        self.loaded_data_symbols: Set[DataSymbol] = set()
        self.saved_store_data: List[SavedStoreData] = []
        self.deep_refs: Set[DeepRef] = set()
        self.mutations: Set[Mutation] = set()
        self.deep_ref_candidates: List[DeepRefCandidate] = []
        self.nested_call_stack: TextualCallNestingStack = []
        self.should_record_args_stack: List[bool] = []
        self.literal_namespace: Optional[NamespaceScope] = None
        self.stack: AttrSubStack = []
        self._waiting_for_call = False

    @property
    def active_scope_for_call(self) -> 'Scope':
        if self._waiting_for_call:
            return self.nested_call_stack[-1][0]
        return self.active_scope

    def __del__(self):
        if hasattr(builtins, self.attrsub_tracer_name):
            delattr(builtins, self.attrsub_tracer_name)
        if hasattr(builtins, self.end_tracer_name):
            delattr(builtins, self.end_tracer_name)
        if hasattr(builtins, self.arg_recorder_name):
            delattr(builtins, self.arg_recorder_name)
        if hasattr(builtins, self.scope_pusher_name):
            delattr(builtins, self.scope_pusher_name)
        if hasattr(builtins, self.scope_popper_name):
            delattr(builtins, self.scope_popper_name)
        if hasattr(builtins, self.literal_tracer_name):
            delattr(builtins, self.literal_tracer_name)

    def push_stack(self, new_scope: 'Scope'):
        self.stack.append((
            self.saved_store_data,
            self.deep_refs,
            self.mutations,
            self.deep_ref_candidates,
            self.active_scope,
            self.original_active_scope,
            self.nested_call_stack,
            self.should_record_args,
            self.should_record_args_stack,
            self.literal_namespace
        ))
        self.saved_store_data = []
        self.deep_refs = set()
        self.mutations = set()
        self.deep_ref_candidates = []
        self.original_active_scope = new_scope
        self.active_scope = new_scope
        self.should_record_args = False
        self.should_record_args_stack = []
        self.nested_call_stack = []
        self.literal_namespace = None

    def pop_stack(self):
        (
            self.saved_store_data,
            self.deep_refs,
            self.mutations,
            self.deep_ref_candidates,
            self.active_scope,
            self.original_active_scope,
            self.nested_call_stack,
            self.should_record_args,
            self.should_record_args_stack,
            self.literal_namespace
        ) = self.stack.pop()

    @staticmethod
    def debug_attribute_tracer(obj, attr, ctx):
        logger.debug('%s attr %s of obj %s', ctx, attr, obj)
        return obj

    def _try_to_resync_obj_ref(self, obj, obj_name):
        if obj_name is None:
            return
        dsym = self.safety.trace_state.prev_trace_stmt_in_cur_frame.scope.lookup_data_symbol_by_name_this_indentation(obj_name)
        if dsym is None:
            return
        # namespace = self.safety.namespaces.get(dsym.cached_obj_id, None)
        # if namespace is not None:
        #     del self.safety.namespaces[dsym.cached_obj_id]
        #     self.safety.namespaces[id(obj)] = namespace
        #     namespace.update_obj_ref(obj)
        # namespace = self.safety.namespaces.get(dsym.obj_id, None)
        # if namespace is not None:
        #     del self.safety.namespaces[dsym.obj_id]
        #     self.safety.namespaces[id(obj)] = namespace
        #     namespace.update_obj_ref(obj)
        # self.safety.aliases[dsym.obj_id].discard(dsym)
        # self.safety.aliases[dsym.cached_obj_id].discard(dsym)
        # self.safety.aliases[id(obj)].add(dsym)
        dsym.update_obj_ref(obj)
        # dsym._refresh_cached_obj()

    def attrsub_tracer(self, obj, attr_or_subscript, is_subscript, ctx, call_context, obj_name=None):
        # print(obj_name, self.safety.trace_state.prev_trace_stmt_in_cur_frame.scope)
        # self._try_to_resync_obj_ref(obj, obj_name)
        # print(obj, attr_or_subscript, is_subscript, ctx)
        should_record_args = False
        try:
            if obj is None:
                return None
            if not isinstance(attr_or_subscript, (str, int)):
                return obj
            obj_id = id(obj)
            scope = self.safety.namespaces.get(obj_id, None)
            # print('%s attr %s of obj %s' % (ctx, attr, obj))
            if scope is None:
                class_scope = self.safety.namespaces.get(id(obj.__class__), None)
                if class_scope is not None and not is_subscript:
                    # print('found class scope %s containing %s' % (class_scope, list(class_scope.all_data_symbols_this_indentation())))
                    scope = class_scope.clone(obj)
                    if obj_name is not None:
                        scope.scope_name = obj_name
                else:
                    # print('no scope for class', obj.__class__)
                    # if self.safety.trace_state.prev_trace_stmt.finished:
                    #     # avoid creating new scopes if we already did this computation
                    #     self.active_scope = None
                    #     return obj
                    try:
                        scope_name = next(iter(self.safety.aliases.get(obj_id, None))).name if obj_name is None else obj_name
                    except (TypeError, StopIteration):
                        scope_name = '<unknown namespace>'
                    scope = NamespaceScope(obj, self.safety, scope_name, parent_scope=None)
                self.safety.namespaces[obj_id] = scope
                # FIXME: brittle strategy for determining parent scope of obj
                if scope.parent_scope is None:
                    if (
                        obj_name is not None and
                        obj_name not in self.safety.trace_state.prev_trace_stmt_in_cur_frame.frame.f_locals
                    ):
                        parent_scope = self.safety.global_scope
                    else:
                        parent_scope = self.active_scope
                    scope.parent_scope = parent_scope

            self.active_scope = scope
            # if scope is None:  # or self.safety.trace_state.prev_trace_stmt.finished:
            #     if ctx in ('Store', 'AugStore'):
            #         self.active_scope = self.original_active_scope
            #     return obj
            if scope is None or self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
                return obj
            elif ctx in ('Store', 'AugStore') and scope is not None:
                self.saved_store_data.append((scope, obj, attr_or_subscript, is_subscript))
                # reset active scope here
                self.active_scope = self.original_active_scope
            if ctx == 'Load':
                # save off event counter and obj_id
                # if event counter didn't change when we process the Call retval, and if the
                # retval is None, this is a likely signal that we have a mutation
                # TODO: this strategy won't work if the arguments themselves lead to traced function calls
                # print('looking for', attr_or_subscript)
                data_sym = scope.lookup_data_symbol_by_name_this_indentation(
                    attr_or_subscript, is_subscript=is_subscript
                )
                if data_sym is None:
                    try:
                        obj_attr_or_sub = self.safety.retrieve_namespace_attr_or_sub(
                            obj, attr_or_subscript, is_subscript
                        )
                        symbol_type = DataSymbolType.SUBSCRIPT if is_subscript else DataSymbolType.DEFAULT
                        data_sym = DataSymbol(
                            attr_or_subscript,
                            symbol_type,
                            obj_attr_or_sub,
                            scope,
                            self.safety,
                            stmt_node=None,
                            parents=None,
                            refresh_cached_obj=True
                        )
                        # this is to prevent refs to the scope object from being considered as stale if we just load it
                        data_sym.defined_cell_num = data_sym.required_cell_num = scope.max_defined_timestamp
                        scope.put(attr_or_subscript, data_sym)
                        # print('put', data_sym, 'in', scope.full_namespace_path)
                        # FIXME: DataSymbols should probably register themselves with the alias manager at creation
                        self.safety.aliases[id(obj_attr_or_sub)].add(data_sym)
                    except:
                        pass
                if call_context:
                    should_record_args = True
                    method_special_case = MethodSpecialCase.normal
                    if isinstance(obj, list) and attr_or_subscript == 'append':
                        method_special_case = MethodSpecialCase.list_append
                    self.deep_ref_candidates.append(
                        ((self.trace_event_counter[0], obj_id, obj_name), method_special_case, set())
                    )
                elif data_sym is not None:
                    # TODO: if we have a.b.c, will this consider a.b loaded as well as a.b.c? This is bad if so.
                    self.loaded_data_symbols.add(data_sym)
            return obj
        finally:
            if call_context:
                self.should_record_args_stack.append(self.should_record_args)
                self.should_record_args = should_record_args

    def end_tracer(self, obj, call_context):
        if self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
            self.active_scope = self.original_active_scope
            return obj
        if call_context and len(self.deep_ref_candidates) > 0:
            (evt_counter, obj_id, obj_name), method_special_case, recorded_args = self.deep_ref_candidates.pop()
            if evt_counter == self.trace_event_counter[0]:
                if obj is None:
                    self.mutations.add((obj_id, tuple(recorded_args), method_special_case))
                else:
                    self.deep_refs.add((obj_id, obj_name, tuple(recorded_args)))
        # print('reset active scope from', self.active_scope, 'to', self.original_active_scope)
        self.active_scope = self.original_active_scope
        return obj

    def arg_recorder(self, obj, name):
        if self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished or not self.should_record_args:
            return obj
        if len(self.deep_ref_candidates) == 0:
            logger.error('Error: no associated symbol for recorded args; skipping recording')
            return obj

        if isinstance(name, str):
            self.deep_ref_candidates[-1][-1].add(name)
        elif isinstance(name, tuple) and len(name) > 0:
            recorded_arg = AttrSubSymbolChain(name)
            self.deep_ref_candidates[-1][-1].add(recorded_arg)

        return obj

    def scope_pusher(self, obj):
        # if self.safety.trace_state.prev_trace_stmt.finished:
        #     return obj
        self.nested_call_stack.append((self.active_scope, self._waiting_for_call))
        self._waiting_for_call = True
        self.active_scope = self.original_active_scope
        return obj

    def scope_popper(self, obj, should_pop_should_record_args_stack):
        # if self.safety.trace_state.prev_trace_stmt.finished:
        #     return obj
        self.active_scope, self._waiting_for_call = self.nested_call_stack.pop()
        if should_pop_should_record_args_stack:
            self.should_record_args = self.should_record_args_stack.pop()
        return obj

    def literal_tracer(self, literal):
        if self.safety.trace_state.prev_trace_stmt_in_cur_frame.finished:
            return literal
        if isinstance(literal, (dict, list, tuple)):
            scope = NamespaceScope(
                literal, self.safety, None, self.safety.trace_state.prev_trace_stmt_in_cur_frame.scope
            )
            gen = literal.items() if isinstance(literal, dict) else enumerate(literal)
            for i, obj in gen:
                scope.upsert_data_symbol_for_name(
                    i, obj, set(), self.safety.trace_state.prev_trace_stmt_in_cur_frame.stmt_node, True
                )
            self.literal_namespace = scope
        return literal

    def stmt_transition_hook(self):
        self._waiting_for_call = False

    def reset(self):
        self.loaded_data_symbols = set()
        self.saved_store_data = []
        self.deep_refs = set()
        self.mutations = set()
        self.deep_ref_candidates = []
        self.active_scope = self.original_active_scope
        self.should_record_args = False
        self.literal_namespace = None
        # self.nested_call_stack = []
        # self.stmt_transition_hook()


class AttrSubTracingNodeTransformer(ast.NodeTransformer):
    def __init__(
            self,
            attrsub_tracer: str,
            end_tracer: str,
            arg_recorder: str,
            scope_pusher: str,
            scope_popper: str,
            literal_tracer: str
    ):
        self.attrsub_tracer = attrsub_tracer
        self.end_tracer = end_tracer
        self.arg_recorder = arg_recorder
        self.scope_pusher = scope_pusher
        self.scope_popper = scope_popper
        self.literal_tracer = literal_tracer
        self.inside_attrsub_load_chain = False
        self.skip_lines: Set[int] = set()

    @contextmanager
    def attrsub_load_context(self, override=True):
        old = self.inside_attrsub_load_chain
        self.inside_attrsub_load_chain = override
        yield
        self.inside_attrsub_load_chain = old

    def visit_Attribute(self, node: 'ast.Attribute', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Subscript(self, node: 'ast.Subscript', call_context=False):
        return self.visit_Attribute_or_Subscript(node, call_context)

    def visit_Attribute_or_Subscript(self, node: 'Union[ast.Attribute, ast.Subscript]', call_context=False):
        is_load = isinstance(node.ctx, ast.Load)
        is_subscript = isinstance(node, ast.Subscript)
        # TODO: expand beyond simple slices
        if is_subscript:
            sub_node = cast(ast.Subscript, node)
            if isinstance(sub_node.slice, ast.Index):
                attr_or_sub = sub_node.slice.value
                # if isinstance(attr_or_sub, ast.Str):
                #     attr_or_sub = attr_or_sub.s
                # elif isinstance(attr_or_sub, ast.Num):
                #     attr_or_sub = attr_or_sub.n
                # else:
                #     logger.debug('unimpled index: %s', attr_or_sub)
                #     return node
            else:
                logger.debug('unimpled slice: %s', sub_node.slice)
                return node
            # elif isinstance(sub_node.slice, ast.Slice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # elif isinstance(sub_node.slice, ast.ExtSlice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # else:
            #     raise ValueError('unexpected slice: %s' % sub_node.slice)
        else:
            attr_node = cast(ast.Attribute, node)
            attr_or_sub = ast.Str(attr_node.attr)

        extra_args = []
        if isinstance(node.value, ast.Name):
            extra_args = [ast.Str(node.value.id)]

        with self.attrsub_load_context():
            replacement_value = ast.Call(
                func=ast.Name(self.attrsub_tracer, ast.Load()),
                args=[
                    self.visit(node.value),
                    attr_or_sub,
                    ast.NameConstant(is_subscript),
                    ast.Str(node.ctx.__class__.__name__),
                    ast.NameConstant(call_context),
                ] + extra_args,
                keywords=[]
            )
        ast.copy_location(replacement_value, node.value)
        node.value = replacement_value
        new_node: Union[ast.Attribute, ast.Subscript, ast.Call] = node
        if not self.inside_attrsub_load_chain and is_load:
            new_node = ast.Call(
                func=ast.Name(self.end_tracer, ast.Load()),
                args=[node, ast.NameConstant(call_context)],
                keywords=[]
            )
        return new_node

    def _get_replacement_args(self, args, should_record, keywords):
        replacement_args = []
        for arg in args:
            if keywords:
                maybe_kwarg = getattr(arg, 'value')
            else:
                maybe_kwarg = arg
            chain = GetAttrSubSymbols()(maybe_kwarg)
            statically_resolvable = []
            for sym in chain.symbols:
                # TODO: only handles attributes properly; subscripts will break
                if not isinstance(sym, str):
                    break
                statically_resolvable.append(ast.Str(sym))
            statically_resolvable = ast.Tuple(elts=statically_resolvable, ctx=ast.Load())
            with self.attrsub_load_context(False):
                visited_maybe_kwarg = self.visit(maybe_kwarg)
            ast.copy_location(visited_maybe_kwarg, maybe_kwarg)
            argrecord_args = [visited_maybe_kwarg, statically_resolvable]
            if should_record:
                with self.attrsub_load_context(False):
                    new_arg_value = cast(ast.expr, ast.Call(
                        func=ast.Name(self.arg_recorder, ast.Load()),
                        args=argrecord_args,
                        keywords=[]
                    ))
            else:
                new_arg_value = visited_maybe_kwarg
            # ast.copy_location(new_arg_value, maybe_kwarg)
            if keywords:
                setattr(arg, 'value', new_arg_value)
            else:
                arg = new_arg_value
            replacement_args.append(arg)
        return replacement_args

    def visit_Call(self, node: ast.Call):
        is_attrsub = False
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            is_attrsub = True
            with self.attrsub_load_context():
                node.func = self.visit_Attribute_or_Subscript(node.func, call_context=True)

            # TODO: need a way to rewrite ast of attribute and subscript args,
            #  and to process these separately from outer rewrite

        node.args = self._get_replacement_args(node.args, is_attrsub, False)
        node.keywords = self._get_replacement_args(node.keywords, is_attrsub, True)

        # in order to ensure that the args are processed with appropriate active scope,
        # we need to push current active scope before processing the args and pop after
        # (pop happens on function return as opposed to in tracer)
        node.func = ast.Call(
            func=ast.Name(self.scope_pusher, ast.Load()),
            args=[node.func],
            keywords=[],
        )

        node = ast.Call(
            func=ast.Name(self.scope_popper, ast.Load()),
            args=[node, ast.NameConstant(is_attrsub)],
            keywords=[]
        )

        if self.inside_attrsub_load_chain or not is_attrsub:
            return node

        replacement_node = ast.Call(
            func=ast.Name(self.end_tracer, ast.Load()),
            args=[node, ast.NameConstant(True)],
            keywords=[]
        )
        ast.copy_location(replacement_node, node)
        return replacement_node

    def visit_Assign(self, node: ast.Assign):
        if not isinstance(node.value, (ast.List, ast.Tuple)):
            return self.generic_visit(node)

        new_targets = []
        for target in node.targets:
            new_targets.append(self.visit(target))
        node.targets = cast('List[ast.expr]', new_targets)
        replacement_literal = ast.Call(
            func=ast.Name(self.literal_tracer, ast.Load()),
            args=[node.value],
            keywords=[]
        )
        ast.copy_location(node.value, replacement_literal)
        node.value = replacement_literal
        return node

    def visit(self, node):
        if getattr(node, 'lineno', -1) in self.skip_lines:
            return node
        else:
            return super().visit(node)



import ast
import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union


class StatementInserter(ast.NodeTransformer):
    def __init__(self, stmt_to_insert_before: 'Union[str, ast.stmt]'):
        if isinstance(stmt_to_insert_before, str):
            self.stmt_to_insert_before = ast.parse(stmt_to_insert_before).body[0]
        else:
            self.stmt_to_insert_before = stmt_to_insert_before

    def visit(self, node):
        if not hasattr(node, 'body'):
            return node
        if not all(isinstance(nd, ast.stmt) for nd in node.body):
            return node
        new_stmts = []
        for stmt in node.body:
            insert_stmt = copy.deepcopy(self.stmt_to_insert_before)
            ast.copy_location(insert_stmt, stmt)
            new_stmts.append(insert_stmt)
            new_stmts.append(self.visit(stmt))
        node.body = new_stmts
        return node



# -*- coding: utf-8 -*-
import ast
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain
from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs

if TYPE_CHECKING:
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import Scope
    from nbsafety.types import SymbolRef
    from typing import Set, Tuple

logger = logging.getLogger(__name__)


def get_symbols_for_references(
        symbol_refs: 'Set[SymbolRef]', scope: 'Scope'
) -> 'Tuple[Set[DataSymbol], Set[DataSymbol]]':
    symbols = set()
    called_symbols = set()
    for symbol_ref in symbol_refs:
        if isinstance(symbol_ref, str):
            dsym = scope.lookup_data_symbol_by_name(symbol_ref)
            called_dsym = None
        elif isinstance(symbol_ref, AttrSubSymbolChain):
            dsym, called_dsym = scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref)
        else:
            logger.warning('invalid type for ref %s', symbol_ref)
            continue
        if dsym is not None:
            symbols.add(dsym)
        if called_dsym is not None:
            called_symbols.add(called_dsym)
    return symbols, called_symbols


def compute_call_chain_live_symbols(live: 'Set[DataSymbol]'):
    seen = set()
    worklist = list(live)
    while len(worklist) > 0:
        called_dsym = worklist.pop()
        if called_dsym in seen:
            continue
        seen.add(called_dsym)
        # TODO: handle callable classes
        if not called_dsym.is_function:
            continue
        live_refs, _ = compute_live_dead_symbol_refs(
            cast(ast.FunctionDef, called_dsym.stmt_node).body, called_dsym.get_call_args()
        )
        live_symbols, called_symbols = get_symbols_for_references(live_refs, called_dsym.call_scope)
        worklist.extend(called_symbols)
        live_symbols = set(sym for sym in live_symbols if sym.is_globally_accessible)
        called_symbols = set(sym for sym in called_symbols if sym.is_globally_accessible)
        live |= live_symbols.union(called_symbols)
    return live



# -*- coding: utf-8 -*-
import ast
# from .utils import skipif_known_failing

from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs


def _remove_callpoints(symbols):
    return set(sym for sym in symbols if isinstance(sym, str))


def test_simple():
    live, dead = compute_live_dead_symbol_refs("""
x = 5
print(foo, x)""")
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'foo', 'print'}
    assert dead == {'x'}


def test_function_body():
    fbody = ast.parse("""
def func():
    y = 42
    print(foo, bar, baz, x)
    x = 5
""").body[0].body
    live, dead = compute_live_dead_symbol_refs(fbody)
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'foo', 'bar', 'baz', 'x', 'print'}
    assert dead == {'x', 'y'}



# -*- coding: utf-8 -*-
from .attrsub_tracing import AttrSubTracingManager, AttrSubTracingNodeTransformer
from .trace_stmt import TraceStatement
from .trace_state import TraceState
from .tracer import make_tracer



# -*- coding: utf-8 -*-
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.ipython_utils import cell_counter

if TYPE_CHECKING:
    from typing import Any, Set
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import NamespaceScope
    from nbsafety.safety import NotebookSafety

logger = logging.getLogger(__name__)
NOT_FOUND = object()


class LegacyUpdateProtocol(object):
    def __init__(self, safety: 'NotebookSafety', updated_dep: 'DataSymbol', mutated: bool):
        self.safety = safety
        self.updated_sym = updated_dep
        self.seen: Set[DataSymbol] = set()
        self.parent_seen: Set[DataSymbol] = set()
        self.mutated = mutated

    def __call__(self, propagate=True):
        if propagate:
            # print(self, 'update deps')
            self._propagate_update(self.updated_sym, self.updated_sym._get_obj(), refresh=True)

    def _propagate_update_to_deps(self, dsym: 'DataSymbol'):
        # print(self, 'propagate to child deps', self.children)
        if dsym.should_mark_stale(self.updated_sym):
            # if self.full_namespace_path == updated_dep.full_namespace_path:
            #     print('weird', self.full_namespace_path, self, updated_dep, self.obj_id, updated_dep.obj_id, self.cached_obj_id, updated_dep.cached_obj_id)
            dsym.fresher_ancestors.add(self.updated_sym)
            dsym.required_cell_num = cell_counter()
        for child in dsym.children:
            # print(self, 'prop to', child, self._get_children_to_skip())
            self._propagate_update(child, child._get_obj())

    def _get_children_to_skip(self, dsym: 'DataSymbol', obj_id=None):
        # TODO: this should probably DFS in order to skip all namespace ancestors
        if obj_id is None:
            obj_id = dsym.obj_id
        children_to_skip = set()
        # skip any children that themselves contain an alias of self
        # TODO: this definitely won't work unless we are aware of containing namespaces
        #  need to therefore create data symbols for list literals and things like lst `append`
        for self_alias in self.safety.aliases[obj_id]:
            containing_obj_id = getattr(self_alias.containing_scope, 'obj_id', None)
            if containing_obj_id is not None:
                children_to_skip |= self.safety.aliases[containing_obj_id]
        return children_to_skip

    def _propagate_update(self, dsym: 'DataSymbol', new_parent_obj: 'Any', refresh=False):
        # look at old obj_id and cur obj_id
        # walk down namespace hierarchy from old obj_id, and track corresponding DCs from cur obj_id
        # a few cases to consider:
        # 1. mismatched obj ids or unavailable from new hierarchy:
        #    mark old dc as mutated AND stale, and propagate to old dc children
        # 2. new / old DataSymbols have same obj ids:
        #    mark old dc as mutated, but NOT stale, and propagate to children
        #    Q: should old dc additionally be refreshed?
        #    Technically it should already be fresh, since if it's still a descendent of this namespace, we probably
        #    had to ref the namespace ancestor, which should have been caught by the checker if the descendent has
        #    some other stale ancestor. If not fresh, let's mark it so and log a warning about a potentially stale usage
        if dsym._tombstone or dsym in self.seen:
            return
        self.seen.add(dsym)
        # print('propagate update from', self)

        new_id = None if new_parent_obj is NOT_FOUND else id(new_parent_obj)

        if self.updated_sym is dsym:
            old_parent_obj = dsym._get_cached_obj()
            old_id = dsym.cached_obj_id
        else:
            old_parent_obj = dsym._get_obj()
            old_id = dsym.obj_id

        namespace = self.safety.namespaces.get(old_id, None)
        if namespace is None and refresh:  # if we are at a leaf
            self._propagate_update_to_namespace_parents(dsym, refresh=refresh)
        if namespace is not None and (old_id != new_id or self.mutated or not refresh):
            for dc in namespace.all_data_symbols_this_indentation(exclude_class=True):
                should_refresh = refresh  # or updated_dep in set(namespace.all_data_symbols_this_indentation())
                dc_in_self_namespace = False
                if new_parent_obj is NOT_FOUND:
                    self._propagate_update(dc, NOT_FOUND, refresh=should_refresh)
                else:
                    try:
                        obj = dsym._get_obj()
                        obj_attr_or_sub = self.safety.retrieve_namespace_attr_or_sub(obj, dc.name, dc.is_subscript)
                        # print(dc, obj, obj_attr_or_sub, updated_dep, seen, parent_seen, refresh, mutated, old_id, new_id)
                        self._propagate_update(dc, obj_attr_or_sub, refresh=should_refresh)
                        dc_in_self_namespace = True
                        if new_parent_obj is not old_parent_obj and new_id is not None:
                            new_namespace = self.safety.namespaces.get(new_id, None)
                            if new_namespace is None:
                                new_namespace = namespace.shallow_clone(new_parent_obj)
                                self.safety.namespaces[new_id] = new_namespace
                            # TODO: handle class data cells properly;
                            #  in fact; we still need to handle aliases of class data cells
                            if dc.name not in new_namespace.data_symbol_by_name(dc.is_subscript):
                                new_dc = dc.shallow_clone(obj_attr_or_sub, new_namespace, dc.symbol_type)
                                new_namespace.put(dc.name, new_dc)
                                self.safety.updated_symbols.add(new_dc)
                    except (KeyError, IndexError, AttributeError):
                        self._propagate_update(dc, NOT_FOUND, refresh=should_refresh)
                if dc_in_self_namespace and dc.should_mark_stale(self.updated_sym):
                    # print(self, 'add', dc, 'to namespace stale symbols due to', updated_dep, self.defined_cell_num, dc.defined_cell_num, updated_dep.defined_cell_num, dc.fresher_ancestors)
                    dsym.namespace_stale_symbols.add(dc)
                else:
                    dsym.namespace_stale_symbols.discard(dc)

        # if mutated or self.cached_obj_id != self.obj_id:
        # if (old_id != new_id or not refresh) and not mutated:
        #     self._propagate_update_to_deps(updated_dep, seen, parent_seen)
        # elif mutated:
        if old_id != new_id or not refresh or self.mutated:
            to_skip = self._get_children_to_skip(dsym)
            # self.seen |= to_skip
            old_seen = self.seen
            self.seen = self.seen | to_skip
            self._propagate_update_to_deps(dsym)
            self.seen = old_seen
            # print(self, 'propagate+skip done')

        if self.updated_sym is dsym:
            return

        if refresh:
            for alias in self.safety.aliases[old_id]:
                if alias.defined_cell_num < alias.required_cell_num < cell_counter():
                    logger.debug('possible stale usage of namespace descendent %s' % alias)
                if len(alias.namespace_stale_symbols) > 0:
                    logger.debug('unexpected stale namespace symbols for symbol %s: %s' % (alias, alias.namespace_stale_symbols))
                    alias.namespace_stale_symbols.clear()
                if old_id != new_id or self.mutated:
                    # TODO: better equality testing
                    #  Doing equality testing properly requires that we still have a reference to old object around;
                    #  we should be using weak refs, which complicates this.
                    #  Depth also makes this challenging.
                    self._propagate_update_to_deps(alias)
                # TODO: this will probably work, but will also unnecessarily propagate the refresh up to
                #  namespace parents. Need a mechanism to mark for refresh without parent propagation.
                self.safety.updated_symbols.add(alias)
        else:
            # propagating to all aliases was causing problems
            # see test `test_class_redef`
            self._propagate_update_to_deps(dsym)
            # for alias in self.safety.aliases[old_id]:
            #     # print('propagate', updated_dep, 'to', alias, 'via', self, updated_dep.defined_cell_num, alias.defined_cell_num, self.defined_cell_num)
            #     alias._propagate_update_to_deps(updated_dep, seen, parent_seen)
            if namespace is None and dsym.should_mark_stale(self.updated_sym):  # if we are at a leaf
                # print('propagate', updated_dep, 'to', self, updated_dep.defined_cell_num, self.defined_cell_num)
                self._propagate_update_to_namespace_parents(dsym, refresh=refresh)
        # print(self, 'done')

    def _propagate_update_to_namespace_parents(self, dsym: 'DataSymbol', refresh):
        if not dsym.containing_scope.is_namespace_scope or dsym in self.parent_seen:
            return
        # print('parent propagate', self)
        self.parent_seen.add(dsym)
        containing_scope = cast('NamespaceScope', dsym.containing_scope)
        containing_namespace_obj_ids_to_consider = {containing_scope.obj_id}
        # if refresh:
        #     for self_alias in self.safety.aliases[self.obj_id]:
        #         if self_alias.containing_scope.is_namespace_scope:
        #             containing_namespace_obj_ids_to_consider.add(getattr(self_alias.containing_scope, 'obj_id', None))
        #     containing_namespace_obj_ids_to_consider.discard(None)
        for containing_namespace_obj_id in containing_namespace_obj_ids_to_consider:
            if containing_namespace_obj_id in self.parent_seen:
                continue
            self.parent_seen.add(containing_namespace_obj_id)
            children_to_skip = self._get_children_to_skip(dsym, obj_id=containing_namespace_obj_id)
            for alias in self.safety.aliases[containing_namespace_obj_id]:
                # print('propagate from ns parent', alias, 'with refresh=', refresh)
                if refresh and not dsym.is_stale:
                    alias.namespace_stale_symbols.discard(dsym)
                    if not alias.is_stale:
                        alias.fresher_ancestors = set()
                self._propagate_update_to_namespace_parents(alias, refresh)
                if not refresh and dsym.should_mark_stale(self.updated_sym):
                    # print(self, 'mark stale due to', updated_dep, 'which is in namespace of', alias)
                    alias.namespace_stale_symbols.add(dsym)
                if refresh:
                    # containing_scope.max_defined_timestamp = cell_counter()
                    self.safety.updated_scopes.add(containing_scope)
                for alias_child in alias.children:
                    if alias_child.obj_id == containing_namespace_obj_id or alias_child in children_to_skip:
                        continue
                    # should_refresh = containing_namespace_obj_id == getattr(alias_child.containing_scope, 'obj_id', None)
                    # if alias_child.obj_id == alias_child.cached_obj_id and alias_child in children_to_skip:
                    #     continue
                    # Next, complicated check to avoid propagating along a class -> instance edge.
                    # The only time this is OK is when we changed the class, which will not be the case here.
                    alias_child_namespace = alias_child.namespace
                    if alias_child_namespace is not None:
                        if alias_child_namespace.cloned_from is containing_scope:
                            if self.updated_sym.namespace is not containing_scope:
                                continue
                    # if len(self.safety.aliases[alias_child.obj_id] & seen) > 0:
                    #     continue
                    # print(self, alias, self.containing_scope, containing_namespace_obj_id, 'propagate to', alias_child, refresh, alias.is_stale, children_to_skip)
                    self._propagate_update(alias_child, alias_child._get_obj())



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List
    from nbsafety.safety import NotebookSafety


# USAGE = """Options:
#
# show_graph:
#     - This will print out the dependency graph of global variables. Stale nodes are labeled red.
#       Notice that user might need to call this twice to have it to work.

USAGE = """Options:

show_dependency <variable_name> <variable_name2> ...: 
    - This will print out the dependencies for given global variables.
      Multiple variables should be separated with spaces.

show_stale: 
    - This will print out all the global variables that are stale. 

remove_dependency <parent_name> <child_name>:
    - This will remove the dependency between parent variable and the child variable.

add_dependency <parent_name> <child_name>:
    - This will add the dependency between parent variable and the child variable.

turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn off the warnings for given global variables. These variables will not be
      considered as stale anymore. Multiple variables should be separated with spaces.

turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn the warnings back on for given global variables. These variables could have
      stale dependencies now. Multiple variables should be separated with spaces."""


def show_deps(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) == 1:
        print("Usage: %safety show_dependency <variable_name> <variable_name2> ...")
        return
    for data_sym_name in line[1:]:
        data_sym = safety.global_scope.lookup_data_symbol_by_name(data_sym_name)
        if data_sym:
            print("DataSymbol {} (defined {}; required {}) is dependent on {}".format(
                data_sym.readable_name,
                data_sym.defined_cell_num,
                data_sym.required_cell_num,
                [str(n) for n in data_sym.parents] if data_sym.parents else "Nothing"
            ))
        else:
            print("Cannot find DataSymbol", data_sym_name)


def show_stale(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) < 2 or line[1] == 'global':
        dsym_sets: Any = [safety.global_scope.all_data_symbols_this_indentation()]
    elif line[1] == 'all':
        dsym_sets = safety.aliases.values()
    else:
        print("TODO: show usage statement")
        return
    stale_set = set()
    for dsym_set in dsym_sets:
        for data_sym in dsym_set:
            if data_sym.is_stale:
                stale_set.add(data_sym)
    if not stale_set:
        print("No DataSymbol has stale dependency for now!")
    elif len(stale_set) == 1:
        print("The only DataSymbol with stale dependencies is:", str(stale_set.pop()))
    else:
        print("DataSymbols with stale dependencies are:", [str(n) for n in stale_set])


def set_propagation(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) != 2 or line[1] not in ['between', 'within']:
        print('Usage: %safety set_propagation [between|within]')
        # TODO: complete explanation
        return
    safety.config.intra_cell_staleness_propagation = (line[1] == 'within')


def trace_messages(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) != 2:
        print("Usage: %safety trace_messages [enabled|disabled] ...")
        return
    safety.config.trace_messages_enabled = (line[1].lower().startswith("enable"))


def remove_dep(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) != 3:
        print("Usage: %safety remove_dependency <parent_name> <child_name>")
        return
    parent_data_sym = safety.global_scope.lookup_data_symbol_by_name(line[1])
    if not parent_data_sym:
        print("Cannot find DataSymbol", line[1])
        return
    child_data_sym = safety.global_scope.lookup_data_symbol_by_name(line[2])
    if not child_data_sym:
        print("Cannot find DataSymbol", line[2])
        return
    if child_data_sym not in parent_data_sym.children or parent_data_sym not in child_data_sym.parents:
        print("Two DataSymbols do not have a dependency relation")
        return
    parent_data_sym.children.remove(child_data_sym)
    child_data_sym.parents.remove(parent_data_sym)


def add_dep(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) != 3:
        print("Usage: %safety add_dependency <parent_name> <child_name>")
        return
    parent_data_sym = safety.global_scope.lookup_data_symbol_by_name(line[1])
    if not parent_data_sym:
        print("Cannot find DataSymbol", line[1])
        return
    child_data_sym = safety.global_scope.lookup_data_symbol_by_name(line[2])
    if not child_data_sym:
        print("Cannot find DataSymbol", line[2])
        return
    if child_data_sym in parent_data_sym.children and parent_data_sym in child_data_sym.parents:
        print("Two DataSymbols already have a dependency relation")
        return
    parent_data_sym.children.add(child_data_sym)
    child_data_sym.parents.add(parent_data_sym)


def turn_off_warnings_for(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) <= 1:
        print("Usage: %safety turn_off_warnings_for <variable_name> <variable_name2> ...")
        return
    for data_sym_name in line[1:]:
        data_sym = safety.global_scope.lookup_data_symbol_by_name(data_sym_name)
        if data_sym:
            data_sym.disable_warnings = True
            print("Warnings are turned off for", data_sym_name)
        else:
            print("Cannot find DataSymbol", data_sym_name)


def turn_on_warnings_for(safety: 'NotebookSafety', line: 'List[str]'):
    if len(line) <= 1:
        print("Usage: %safety turn_on_warnings_for <variable_name> <variable_name2> ...")
        return
    for data_sym_name in line[1:]:
        data_sym = safety.global_scope.lookup_data_symbol_by_name(data_sym_name)
        if data_sym:
            data_sym.disable_warnings = False
            print("Warnings are turned on for", data_sym_name)
        else:
            print("Cannot find DataSymbol", data_sym_name)



# -*- coding: utf-8 -*-
import itertools

from nbsafety.analysis.stmt_edges import get_statement_symbol_edges
from nbsafety.analysis.lineno_stmt_map import compute_lineno_to_stmt_mapping

from .utils import skipif_known_failing


def get_statement_lval_and_rval_symbols(node):
    edges, _ = get_statement_symbol_edges(node)
    lvals = set(edges.keys())
    if len(edges) == 0:
        rvals = set()
    else:
        rvals = set.union(*edges.values())
    return lvals - {None}, rvals - {None}


def get_directed_edge_list(node):
    edges, _ = get_statement_symbol_edges(node)
    edge_list = []
    for k, v in edges.items():
        if k is None:
            continue
        vset = set(v) - {None}
        edge_list.extend((k, val) for val in vset)
    return sorted(edge_list)


def test_classes():
    code = """
class Foo(object):
    pass
    
class Bar(Foo):
    pass
    
class Baz(Foo, Bar):
    pass
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'Foo'}
    assert rvals == {'object'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'Bar'}
    assert rvals == {'Foo'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[7])
    assert lvals == {'Baz'}
    assert rvals == {'Foo', 'Bar'}


def test_for_loop():
    code = """
for i in range(10):
    a = i
    b = a + i
    lst = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'i'}
    assert rvals == {'range'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'a'}
    assert rvals == {'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'b'}
    assert rvals == {'a', 'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'lst'}
    assert rvals == {'a', 'b'}


def test_context_manager():
    code = """
fname = 'file.txt'
with open(fname) as f:
    contents = f.read()
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'f'}
    assert rvals == {'fname', 'open'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'contents'}
    # attributes should be skipped
    assert rvals == set()


@skipif_known_failing
def test_list_comp():
    code = 'x = iter([tuple(islice(itr,i,i+n,1)) for i in range(len(itr)-n+1)])'
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'x'}
    assert rvals == {'tuple', 'islice', 'itr', 'n'}
    code = 'x = [i for i in range(10)]'
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'x'}
    assert rvals == {'range'}
    code = 'x = [i for j in range(10)]'
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'x'}
    assert rvals == {'i', 'range'}


def test_deeply_nested_arguments():
    mapping = compute_lineno_to_stmt_mapping('x = f(f(f(f(y, z=7), z=9), z=w), 10)')
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'x'}
    assert rvals == {'f', 'y', 'w'}


def test_unpacked_from_function():
    mapping = compute_lineno_to_stmt_mapping('x, y, z = f(a, b, c, d=e)')
    edges = get_directed_edge_list(mapping[1])
    assert sorted(edges) == sorted(itertools.product(['x', 'y', 'z'], ['a', 'b', 'c', 'e', 'f']))


def test_unpacking_attribution():
    mapping = compute_lineno_to_stmt_mapping('x, y = a, (b, c)')
    edges = get_directed_edge_list(mapping[1])
    assert sorted(edges) == [('x', 'a'), ('y', 'b'), ('y', 'c')]
    mapping = compute_lineno_to_stmt_mapping('x, (y, z) = a, (b, c)')
    edges = get_directed_edge_list(mapping[1])
    assert sorted(edges) == [('x', 'a'), ('y', 'b'), ('z', 'c')]
    mapping = compute_lineno_to_stmt_mapping('(x1, x2), (y, z) = a, (b, c)')
    edges = get_directed_edge_list(mapping[1])
    assert sorted(edges) == [('x1', 'a'), ('x2', 'a'), ('y', 'b'), ('z', 'c')]


def test_deeply_nested_unpacking_attribution():
    mapping = compute_lineno_to_stmt_mapping('[x1, (x2, x3)], y = ([a11, (a12, a13)], ([a21, a22], a3)), (b, c)')
    edges = get_directed_edge_list(mapping[1])
    assert sorted(edges) == [
        ('x1', 'a11'), ('x1', 'a12'), ('x1', 'a13'),
        ('x2', 'a21'), ('x2', 'a22'), ('x3', 'a3'),
        ('y', 'b'), ('y', 'c')
    ]



# -*- coding: utf-8 -*-
import ast
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import get_attrsub_symbol_chain, AttrSubSymbolChain, CallPoint
from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import List, Optional, Set, Tuple, Union
    from ..types import SymbolRef
    Killable = Union[str, AttrSubSymbolChain]

logger = logging.getLogger(__name__)


# TODO: have the logger warnings additionally raise exceptions for tests
class ComputeLiveSymbolRefs(SaveOffAttributesMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self, init_killed: 'Optional[Set[str]]' = None):
        if init_killed is None:
            self.dead: 'Set[Killable]' = set()
        else:
            self.dead = cast('Set[Killable]', init_killed)
        self.in_kill_context = False

    def __call__(self, module_node: ast.Module):
        """
        This function should be called when we want to do a liveness check on a
        cell's corresponding ast.Module. For each line/block of the cell we
        first run the check of new assignments, then we obtain all the names.
        In these names, we put the ones that are user defined and not in the
        killed set into the return check_set for further checks.
        """
        # TODO: this will break if we ref a variable in a loop before killing it in the
        #   same loop, since we will add everything on the LHS of an assignment to the killed
        #   set before checking the loop body for live variables
        live = set()
        for node in module_node.body:
            self.visit(node)
            for ref in _get_all_symbol_refs(node):
                if ref in self.dead:
                    continue
                # TODO: check for all subchains in the safe set, not just the first symbol
                if isinstance(ref, AttrSubSymbolChain):
                    if len(ref.symbols) == 0:
                        # can happen if user made syntax error like [1, 2, 3][4, 5, 6] (e.g. forgot comma)
                        continue
                    leading_symbol = ref.symbols[0]
                    if isinstance(leading_symbol, str) and leading_symbol in self.dead:
                        continue
                    if isinstance(leading_symbol, CallPoint) and leading_symbol.symbol in self.dead:
                        continue
                live.add(ref)
        # print(self.safe_set)
        # print(check_set)
        return live, self.dead

    def kill_context(self):
        return self.push_attributes(in_kill_context=True)

    # In case of assignment, we put the new assigned variable into a safe_set
    # to indicate that we know for sure it won't have stale dependency.  Note
    # that node.targets might contain multiple ast.Name node in the case of
    # "a = b = 3", so we go through each node in the targets.  Also note that
    # `target` would be an ast.Tuple node in the case of "a,b = 3,4". Thus
    # we need to break the tuple in that case.
    def visit_Assign(self, node: ast.Assign):
        with self.kill_context():
            self.generic_visit(node.targets)
        self.visit(node.value)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        self.visit_Assign_or_AugAssign_target(node.target)
        self.visit(node.value)

    def visit_AugAssign(self, node: ast.AugAssign):
        self.visit_Assign_or_AugAssign_target(node.target)
        self.visit(node.value)

    def visit_Assign_or_AugAssign_target(self, target_node: 'Union[ast.Attribute, ast.Name, ast.Subscript, ast.expr]'):
        if isinstance(target_node, ast.Name):
            self.dead.add(target_node.id)
        elif isinstance(target_node, (ast.Attribute, ast.Subscript)):
            self.dead.add(get_attrsub_symbol_chain(target_node))
        else:
            logger.warning('unsupported type for node %s' % target_node)

    # We also put the name of new functions in the safe_set
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.dead.add(node.name)
        # with self.kill_context():
        #     self.visit(node.args)

    def visit_Name(self, node):
        if self.in_kill_context:
            self.dead.add(node.id)

    def visit_Tuple_or_List(self, node):
        for elt in node.elts:
            self.visit(elt)

    def visit_List(self, node):
        self.visit_Tuple_or_List(node)

    def visit_Tuple(self, node):
        self.visit_Tuple_or_List(node)

    def visit_For(self, node: ast.For):
        # Case "for a,b in something: "
        with self.kill_context():
            self.visit(node.target)

        # Then we keep doing the visit for the body of the loop.
        for line in node.body:
            self.visit(line)

    def visit_GeneratorExp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_DictComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_ListComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_SetComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(self, node):
        # TODO: as w/ for loop, this will have false positives on later live references
        with self.kill_context():
            for gen in node.generators:
                self.visit(gen.target)

    def visit_Lambda(self, node):
        with self.kill_context():
            self.visit(node.args)

    def visit_arg(self, node):
        if self.in_kill_context:
            self.dead.add(node.arg)


def compute_live_dead_symbol_refs(
        code: 'Union[ast.Module, List[ast.stmt], str]',
        init_killed: 'Optional[Set[str]]' = None
) -> 'Tuple[Set[SymbolRef], Set[SymbolRef]]':
    if init_killed is None:
        init_killed = set()
    if isinstance(code, str):
        code = ast.parse(code)
    elif isinstance(code, list):
        code = ast.Module(body=code)
    return ComputeLiveSymbolRefs(init_killed)(code)


# Call GetAllNames()(ast_tree) to get a set of all names appeared in ast_tree.
# Helper Class
class GetAllSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        self.ref_set: Set[SymbolRef] = set()
        self.inside_attrsub = False
        self.skip_simple_names = False

    def __call__(self, node: ast.AST):
        self.visit(node)
        return self.ref_set

    def attrsub_context(self, inside=True):
        return self.push_attributes(inside_attrsub=inside, skip_simple_names=inside)

    def args_context(self):
        return self.push_attributes(skip_simple_names=False)

    def visit_Name(self, node: ast.Name):
        if not self.skip_simple_names:
            self.ref_set.add(node.id)

    # We overwrite FunctionDef because we don't need to check names in the body of the definition.
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.generic_visit(node.args.defaults)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)

    def visit_Call(self, node: ast.Call):
        with self.args_context():
            self.generic_visit(node.args)
            for kwarg in node.keywords:
                self.visit(kwarg.value)
        self.ref_set.add(get_attrsub_symbol_chain(node))
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            with self.attrsub_context():
                self.visit(node.func)
        else:
            self.visit(node.func)

    def visit_Attribute(self, node: ast.Attribute):
        if not self.inside_attrsub:
            self.ref_set.add(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)

    def visit_Subscript(self, node: ast.Subscript):
        if not self.inside_attrsub:
            self.ref_set.add(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)
        with self.attrsub_context(inside=False):
            self.visit(node.slice)


def _get_all_symbol_refs(node: ast.AST):
    return GetAllSymbolRefs()(node)
